@@ -1,6 +1,6 @@
 #!/usr/bin/perl -w
 
-# (C) Copyright 2010, met.no
+# (C) Copyright 2010-2016 MET Norway
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
@@ -33,12 +33,14 @@ use constant DEFAULT_CTABLE => 'C0000000000000023000';
 my %option = ();
 GetOptions(
            \%option,
+           'ahl=s',        # Decode BUFR messages with AHL matching <ahl_regexp> only
            'all_operators',# Show all operator descriptors when printing section 4
            'bitmap',       # Display bit-mapped values on same line
            'codetables',   # Use code and flag tables to resolve values
            'data_only',    # Print section 4 (data section) only
            'filter=s',     # Decode observations meeting criteria in <filter file> only
            'help',         # Print help information and exit
+           'nodata',       # Do not print (nor decode) section 4 (data section)
            'noqc',         # Do not decode quality control
            'on_error_stop', # Stop processing if an error occurs
            'optional_section',  # Display a hex dump of optional section if present
@@ -59,6 +61,9 @@ pod2usage(-verbose => 0) unless @ARGV;
 # Set verbosity level
 Geo::BUFR->set_verbose($option{verbose}) if $option{verbose};
 
+# Set whether section 4 should be decoded for the BUFR module
+Geo::BUFR->set_nodata() if ($option{nodata});
+
 # Set whether quality information should be decoded for the BUFR module
 Geo::BUFR->set_noqc() if ($option{noqc});
 
@@ -78,6 +83,14 @@ if ($option{tablepath}) {
     Geo::BUFR->set_tablepath(DEFAULT_TABLE_PATH);
 }
 
+my $ahl_regexp;
+if ($option{ahl}) {
+    eval { $ahl_regexp = qr/$option{ahl}/ };
+    die "Argument to --ahl is not a valid Perl regular expression: $@" if $@;
+    # When filtering on ahl we assume file is composed of GTS bulletins only
+    Geo::BUFR->reuse_current_ahl();
+}
+
 # Where to direct output (including verbose output, but not output to STDERR)
 my $OUT;
 if ($option{outfile}) {
@@ -112,6 +125,7 @@ my $sec2_code_ref = sub {return '    Hex dump:'.' 'x26 . unpack('H*',substr(shif
 # Loop for processing of BUFR input files
 foreach my $inputfname ( @ARGV ) {
     my $bufr = Geo::BUFR->new();
+    $bufr->set_filter_cb(\&filter_on_ahl,$ahl_regexp) if $option{ahl};
 
     # Open BUFR file
     $bufr->fopen($inputfname);
@@ -137,19 +151,23 @@ sub decode {
 
         # Read next observation. If an error is encountered during
         # decoding, skip this observation while printing the error
-        # message to STDERR, also displaying ahl of bulletin if found.
+        # message to STDERR, also displaying ahl of bulletin if found
+        # (but skip error message if the message should be skipped on
+        # --ahl anyway).
         my ($data, $descriptors);
         eval {
             ($data, $descriptors) = $bufr->next_observation();
         };
         if ($@) {
+            $current_ahl = $bufr->get_current_ahl() || '';
+            next READLOOP if $option{ahl} && $current_ahl !~ $ahl_regexp; 
+
             warn $@;
             # Try to extract message number and ahl of the bulletin
             # where the error occurred
             $current_message_number = $bufr->get_current_message_number();
             if (defined $current_message_number) {
                 my $error_msg = "In message $current_message_number";
-                $current_ahl = $bufr->get_current_ahl();
                 $error_msg .= " contained in bulletin with ahl $current_ahl\n"
                     if $current_ahl;
                 warn $error_msg if $error_msg;
@@ -157,10 +175,10 @@ sub decode {
             exit(1) if $option{on_error_stop};
             next READLOOP;
         }
-        # Would really liked to test for !$data here, but then we
-        # would miss messages with 0 subsets (which have no data)
 
-        if ($option{codetables}) {
+        next if $option{ahl} && $bufr->is_filtered();
+
+        if ($option{codetables} && !$option{nodata}) {
             # Load C table, trying first to use same table version as
             # the B and D tables loaded in next_observation, or if
             # this C table file does not exist, loads DEFAULT_CTABLE
@@ -174,7 +192,7 @@ sub decode {
         # should have been set to at least 1 (even in a 0 subset message)
         last READLOOP if $current_subset_number == 0;
 
-        if ($current_subset_number == 1) {
+        if ($current_subset_number == 1 || $option{nodata}) {
             $current_message_number = $bufr->get_current_message_number();
             $current_ahl = $bufr->get_current_ahl() || '';
             $message_header = sprintf "\nMessage %d", $current_message_number;
@@ -195,6 +213,7 @@ sub decode {
                 print $OUT $bufr->dumpsection3();
                 $section013_dumped = 1;
             }
+            next READLOOP if $option{nodata};
         } else { # subset number > 1
             next READLOOP if ($option{filter}
                 && filter_observation($bufr, $data, $descriptors));
@@ -281,6 +300,13 @@ sub param {
 
 # Filter routines
 
+sub filter_on_ahl {
+    my $obj = shift;
+    my $ahl_regexp = shift;
+    my $ahl = $obj->get_current_ahl() || '';
+    return $ahl =~ $ahl_regexp ? 0 : 1;
+}
+
 # Read in contents of $filter_file into variables @fid, @fiv,
 # @num_desc, @num_val and $num_criteria, which are defined above.
 sub read_filter_file {
@@ -413,24 +439,28 @@ sub filter_observation {
 
 =pod
 
+=encoding utf8
+
 =head1 SYNOPSIS
 
   bufrread.pl <bufr file(s)>
+      [--ahl <ahl_regexp>]
+      [--all_operators]
+      [--bitmap]
       [--codetables]
       [--data_only]
-      [--param <descriptor file>]
       [--filter <filter file>]
-      [--bitmap]
+      [--help]
+      [--nodata]
       [--noqc]
-      [--outfile <filename>]
-      [--optional_section]
-      [--width n]
-      [--strict_checking n]
       [--on_error_stop]
-      [--all_operators]
+      [--optional_section]
+      [--outfile <filename>]
+      [--param <descriptor file>]
+      [--strict_checking n]
       [--tablepath <path to BUFR tables>]
       [--verbose n]
-      [--help]
+      [--width n]
 
 =head1 DESCRIPTION
 
@@ -438,52 +468,61 @@ Extract BUFR messages from BUFR file(s) and print the decoded content
 to screen, including AHL (Abbreviated Header Line) if present.
 
 Execute without arguments for Usage, with option C<--help> for some
-additional info. See also L</https://wiki.met.no/bufr.pm/start> for
+additional info. See also L<https://wiki.met.no/bufr.pm/start> for
 examples of use.
 
 
 =head1 OPTIONS
 
+   --ahl <ahl_regexp>
+                   Decode BUFR messages with AHL matching <ahl_regexp> only
+   --all_operators Show all operator descriptors when printing section 4
+   --bitmap        Display bit-mapped values on same line
    --codetables    Use code and flag tables to resolve values when unit
                    is [CODE TABLE] or [FLAG TABLE]
    --data_only     Print section 4 (data section) only
-   --param <descriptor file>
-                   Display parameters with descriptors in <descriptor file> only
    --filter <filter file>
                    Decode observations meeting criteria in <filter file> only
-   --bitmap        Display bit-mapped values on same line
+   --help          Display Usage and explain the options used. For even
+                   more info you might prefer to consult perldoc bufrread.pl
+   --nodata        Do not print (nor decode) section 4 (data section)
    --noqc          Do not decode quality control
                    (or any descriptors following 222000)
+   --on_error_stop Stop processing as soon as an error occurs during decoding
    --outfile <filename>
                    Will print to <filename> instead of STDOUT
    --optional_section
                    Display a hex dump of optional section if present
-   --width n       Set width of field used for data values to n characters
-                   (default is 15)
+   --param <descriptor file>
+                   Display parameters with descriptors in <descriptor file> only
    --strict_checking n n=0 (default) Disable strict checking of BUFR format
                        n=1 Issue warning if (recoverable) error in
                            BUFR format
                        n=2 Croak if (recoverable) error in BUFR format.
                            Nothing more in this message/subset will be decoded.
-   --on_error_stop Stop processing as soon as an error occurs during decoding
-   --all_operators Show all operator descriptors when printing section 4
    --tablepath <path to BUFR tables>
                    Set path to BUFR tables (overrides ENV{BUFR_TABLES})
    --verbose n     Set verbose level to n, 0<=n<=6 (default 0). n=1 will
                    show the tables loaded.
-   --help          Display Usage and explain the options used. For even
-                   more info you might prefer to consult perldoc bufrread.pl
+   --width n       Set width of field used for data values to n characters
+                   (default is 15)
 
 Options may be abbreviated, e.g. C<--h> or C<-h> for C<--help>.
 
 To avoid having to use the C<--tablepath> option, you are adviced to
-set the invironment variable BUFR_TABLES to the directory where your
+set the environment variable BUFR_TABLES to the directory where your
 BUFR tables are located (unless the default path provided by
 bufrread.pl works for you).
 
-Each line in <descriptor file> should start with a BUFR descriptor (6
-digits).  Rest of line will be ignored. bufrread.pl will display values
-for these descriptors only.
+For option C<--ahl> the <ahl_regexp> should be a Perl regular
+expression. E.g. C<--ahl 'ISS... ENMI'> will decode only BUFR SHIP
+(ISS) from CCCC=ENMI. This is the only case where a little knowledge
+of Perl might be required for using the utility programs included in
+Geo::BUFR.
+
+For option C<--param> each line in <descriptor file> should start with
+a BUFR descriptor (6 digits).  Rest of line will be ignored.
+bufrread.pl will display values for these descriptors only.
 
 Using C<--filter> will decode only those observations that meet one of
 the criteria in <filter file> (and all of those criteria marked
@@ -503,18 +542,19 @@ file is
 
 which decodes all observations with block number 01, two other
 specific WMO stations and one specific ship, all of which having hour
-(004004) equal to 6 or 7.  If there is no value line after a
+(004004) equal to 6 or 7. If there is no value line after a
 descriptor line, it is enough that the observation contains the
 descriptor(s), whatever the values are. So to extract all ship
-messages from a BUFR SYNOP file, the filter file should contain this
-single line only:
+messages from a BUFR file, the filter file should contain this single
+line only:
 
   D: 001011
 
 If an error occurs during decoding (typically because the required
-BUFR table is missing or message is corrupt) the message is skipped,
-and the number of errors is reported at end of output. You can change
-this default behaviour, however, by setting C<--on_error_stop>.
+BUFR table is missing or message is corrupt), the BUFR message is
+skipped with an error message printed to STDERR, and processing then
+continues with the next BUFR message. You can change this default
+behaviour, however, by setting C<--on_error_stop>.
 
 =head1 CAVEAT
 
@@ -531,6 +571,6 @@ Pl Sannes E<lt>pal.sannes@met.noE<gt>
 
 =head1 COPYRIGHT
 
-Copyright (C) 2010 met.no
+Copyright (C) 2010-2016 MET Norway
 
 =cut
